/**
 * rxxfetch
 * Observable HTTP fetch API for browser and Node.js. Handle 302/303 redirect correctly on Node.js
 *
 * @version 12.5.1
 * @author waiting
 * @license MIT
 * @link https://waitingsong.github.io/rxxfetch/
 */

import { throwError, of, defer } from 'rxjs';
import { catchError, map, concatMap, timeout } from 'rxjs/operators';
import { abortableFetch, AbortController as AbortController$1 } from 'abortcontroller-polyfill/dist/cjs-ponyfill.js';
import _fetch$1, { Headers as Headers$1 } from 'node-fetch';
import NodeFormData from 'form-data';
import QueryString from 'qs';

/* eslint-disable @typescript-eslint/no-unsafe-call */
const initialRxRequestInit = {
    dataType: 'json',
    keepRedirectCookies: false,
    method: 'GET',
    processData: true,
    timeout: null,
    cache: 'default',
    credentials: 'same-origin',
    mode: 'cors',
    redirect: 'follow',
    referrer: 'client',
};
// for node.js
if (typeof window === 'undefined') {
    if (typeof fetch !== 'function') {
        initialRxRequestInit.fetchModule = abortableFetch(_fetch$1).fetch;
    }
    if (typeof Headers !== 'function') {
        initialRxRequestInit.headersInitClass = Headers$1;
    }
}
const httpErrorMsgPrefix = 'Fetch error status:';

function assertNeverRx(x) {
    // eslint-disable-next-line @typescript-eslint/restrict-plus-operands
    return throwError(() => new Error('Assert Never Unexpected object: ' + x));
}

function handleResponseError(resp, bare = false) {
    /* istanbul ignore else */
    if (resp.ok || bare) {
        return of(resp);
    }
    const { status, statusText } = resp;
    return defer(() => resp.text()).pipe(catchError((err) => of(JSON.stringify(err))), map((txt) => {
        const str = `${httpErrorMsgPrefix}${status}\nstatusText: ${statusText}\nResponse: ${txt}`;
        throw new Error(str);
    }));
}
function parseResponseType(response, dataType) {
    let ret;
    switch (dataType) {
        case 'arrayBuffer':
            ret = defer(() => response.arrayBuffer());
            break;
        case 'bare':
            ret = of(response);
            break;
        case 'blob':
            ret = defer(() => response.blob());
            break;
        case 'formData':
            ret = defer(() => response.formData());
            break;
        case 'json':
            // eslint-disable-next-line @typescript-eslint/no-unsafe-assignment
            ret = defer(() => response.json());
            break;
        case 'raw':
            ret = of(response);
            break;
        case 'text':
            ret = defer(() => response.text());
            break;
        default:
            return assertNeverRx(dataType);
    }
    return ret;
}
/** "foo=cookiefoo; Secure; Path=/" */
function parseRespCookie(cookie) {
    /* istanbul ignore else  */
    if (!cookie) {
        return;
    }
    // eslint-disable-next-line require-unicode-regexp
    const arr = cookie.split(/;/);
    const ret = {};
    for (let row of arr) {
        row = row.trim();
        /* istanbul ignore else  */
        if (!row) {
            continue;
        }
        if (!row.includes('=')) {
            continue;
        }
        if (row.startsWith('Path=')) {
            continue;
        }
        const [key, value] = row.split('=');
        /* istanbul ignore else  */
        if (key && value) {
            ret[key] = value;
        }
    }
    /* istanbul ignore else  */
    // eslint-disable-next-line @typescript-eslint/no-unnecessary-condition
    if (ret && Object.keys(ret).length) {
        return ret;
    }
}

var ContentTypeList;
(function (ContentTypeList) {
    ContentTypeList["formDataPartial"] = "multipart/form-data; boundary=";
    ContentTypeList["formUrlencoded"] = "application/x-www-form-urlencoded; charset=utf-8";
    ContentTypeList["html"] = "text/html; charset=utf-8";
    ContentTypeList["json"] = "application/json; charset=utf-8";
    ContentTypeList["plain"] = "text/plain";
})(ContentTypeList || (ContentTypeList = {}));

/* eslint-disable @typescript-eslint/no-unsafe-call */
function buildQueryString(url, data) {
    /* istanbul ignore else */
    if (data && typeof data === 'object' && Object.keys(data).length) {
        const ps = QueryString.stringify(data);
        return url.includes('?') ? `${url}&${ps}` : `${url}?${ps}`;
    }
    return url;
}
/** Split RxRequestInit object to RequestInit and Args */
function splitInitArgs(rxInitOpts) {
    const tmpObj = Object.assign({}, rxInitOpts);
    const args = {};
    /* istanbul ignore else */
    if (typeof tmpObj.cookies !== 'undefined') {
        args.cookies = tmpObj.cookies;
    }
    delete tmpObj.cookies;
    // eslint-disable-next-line @typescript-eslint/unbound-method
    if (tmpObj.abortController && typeof tmpObj.abortController.abort === 'function') {
        args.abortController = tmpObj.abortController;
    }
    delete tmpObj.abortController;
    /* istanbul ignore else */
    if (typeof tmpObj.contentType !== 'undefined') {
        args.contentType = tmpObj.contentType;
    }
    delete tmpObj.contentType;
    /* istanbul ignore else */
    if (typeof tmpObj.data !== 'undefined') {
        args.data = tmpObj.data;
    }
    delete tmpObj.data;
    /* istanbul ignore else */
    if (tmpObj.dataType) {
        args.dataType = tmpObj.dataType;
    }
    delete tmpObj.dataType;
    /* istanbul ignore else */
    if (tmpObj.fetchModule) {
        args.fetchModule = tmpObj.fetchModule;
    }
    delete tmpObj.fetchModule;
    /* istanbul ignore else */
    if (tmpObj.headersInitClass) {
        args.headersInitClass = tmpObj.headersInitClass;
    }
    delete tmpObj.headersInitClass;
    /* istanbul ignore else */
    if (typeof tmpObj.keepRedirectCookies !== 'undefined') {
        args.keepRedirectCookies = !!tmpObj.keepRedirectCookies;
    }
    delete tmpObj.keepRedirectCookies;
    /* istanbul ignore else */
    if (typeof tmpObj.processData !== 'undefined') {
        args.processData = tmpObj.processData;
    }
    delete tmpObj.processData;
    /* istanbul ignore else */
    if (typeof tmpObj.timeout !== 'undefined') {
        args.timeout = tmpObj.timeout;
    }
    delete tmpObj.timeout;
    return {
        args,
        requestInit: Object.assign({}, tmpObj),
    };
}
function parseInitOpts(options) {
    let opts = Object.assign({}, options);
    opts = parseHeaders(opts); // at first!
    opts = parseAbortController(opts);
    opts = parseCookies(opts);
    opts = parseMethod(opts);
    opts.args.dataType = parseDataType(opts.args.dataType);
    opts.args.timeout = parseTimeout$1(opts.args.timeout);
    opts.requestInit.redirect = parseRedirect(opts.args.keepRedirectCookies, opts.requestInit.redirect);
    return opts;
}
function parseHeaders(options) {
    const { args, requestInit } = options;
    /* istanbul ignore else */
    if (args.headersInitClass) { // node.js need pass headers class
        const headers = requestInit.headers
            ? new args.headersInitClass(requestInit.headers)
            : new args.headersInitClass();
        requestInit.headers = headers;
    }
    else if (typeof Headers === 'function') { // browser native
        requestInit.headers = requestInit.headers
            ? new Headers(requestInit.headers)
            : new Headers();
    }
    else {
        throw new TypeError(`parseHeaders(): Headers Class undefined.
If running under Node.js, it must pass HeaderClass such come from package "node-fetch".`);
    }
    const { headers } = requestInit;
    /* istanbul ignore else */
    if (!headers.has('Accept')) {
        headers.set('Accept', 'application/json, text/html, text/javascript, text/plain, */*');
    }
    return { args, requestInit };
}
function parseAbortController(options) {
    const { args, requestInit } = options;
    /* istanbul ignore else */
    // eslint-disable-next-line @typescript-eslint/unbound-method
    // eslint-disable-next-line @typescript-eslint/no-unnecessary-condition
    if (!args.abortController || !args.abortController.signal || typeof args.abortController.abort !== 'function') {
        args.abortController = typeof AbortController === 'function'
            ? new AbortController()
            : new AbortController$1();
    }
    /* istanbul ignore else */
    if (args.abortController) {
        requestInit.signal = args.abortController.signal;
    }
    return { args, requestInit };
}
function parseCookies(options) {
    const { args, requestInit } = options;
    const data = args.cookies;
    const arr = [];
    if (data && typeof data === 'object') {
        for (let [key, value] of Object.entries(data)) {
            /* istanbul ignore else */
            if (key && typeof key === 'string') {
                key = key.trim();
                /* istanbul ignore else */
                if (!key) {
                    continue;
                }
                value = typeof value === 'string' || typeof value === 'number' ? value.toString().trim() : '';
                arr.push(`${key}=${value}`);
            }
        }
    }
    if (arr.length) {
        const headers = requestInit.headers;
        let cookies = headers.get('Cookie');
        if (cookies) {
            cookies = cookies.trim();
            let ret = arr.join('; ');
            /* istanbul ignore if */
            if (cookies.endsWith(';')) {
                cookies = cookies.slice(0, -1);
                ret = `${cookies}; ` + ret;
            }
            else {
                ret = `${cookies}; ` + ret;
            }
            headers.set('Cookie', ret);
        }
        else {
            headers.set('Cookie', arr.join('; '));
        }
    }
    return { args, requestInit };
}
function parseMethod(options) {
    const { args, requestInit } = options;
    if (requestInit.method && ['DELETE', 'POST', 'PUT'].includes(requestInit.method)) {
        const headers = requestInit.headers;
        if (args.contentType === false) ;
        else if (args.contentType) {
            headers.set('Content-Type', args.contentType);
        }
        /* istanbul ignore else */
        else if (!headers.has('Content-Type')) {
            headers.set('Content-Type', ContentTypeList.json);
        }
    }
    return { args, requestInit };
}
/**
 * Parse type of return data
 * @returns default: json
 **/
function parseDataType(value) {
    /* istanbul ignore else */
    if (typeof value === 'string' && ['arrayBuffer', 'bare', 'blob', 'formData', 'json', 'text', 'raw'].includes(value)) {
        return value;
    }
    return 'json';
}
function parseTimeout$1(ps) {
    const value = typeof ps === 'number' && ps >= 0 ? Math.ceil(ps) : null;
    return value === null || !Number.isSafeInteger(value) ? null : value;
}
/**
 * set redirect to 'manul' for retrieve cookies during 301/302 when keepRedirectCookies:TRUE
 * and current value only when "follow"
 */
function parseRedirect(keepRedirectCookies, curValue) {
    // not change value if on Browser
    /* istanbul ignore else */
    if (keepRedirectCookies === true && typeof window === 'undefined') {
        /* istanbul ignore else */
        if (curValue === 'follow') {
            return 'manual';
        }
    }
    return curValue ? curValue : 'follow';
}
/** Select fetch instance from args.fetchModule or native */
function selectFecthModule(mod) {
    let fetchModule = null;
    if (mod) {
        /* istanbul ignore else */
        if (typeof mod !== 'function') {
            throwError(() => new TypeError('fetchModule is not Function'));
        }
        fetchModule = mod;
    }
    /* istanbul ignore else */
    else if (typeof fetch === 'function') { // native fetch
        fetchModule = fetch;
    }
    else {
        throwError(() => new TypeError('fetchModule/fetch not Function'));
    }
    return fetchModule;
}
/**
 * Return input url string
 */
function parseRequestGetLikeData(input, args) {
    let url = '';
    if (typeof args.data === 'undefined') {
        url = input;
    }
    else if (args.processData) { // override the value of body by args.data
        url = buildQueryString(input, args.data);
    }
    else {
        throw new TypeError('Typeof args.data invalid for GET/DELETE when args.processData not true, type is :' + typeof args.data);
    }
    return url;
}
function parseRequestPostLikeData(args) {
    let body;
    const { data } = args;
    if (typeof data === 'string') {
        body = data;
    }
    else if (typeof data === 'undefined' || data === null) ;
    else if (typeof FormData !== 'undefined' && data instanceof FormData) {
        body = data;
    }
    else if (typeof NodeFormData !== 'undefined' && data instanceof NodeFormData) {
        // @ts-expect-error
        body = data;
    }
    else if (typeof Blob !== 'undefined' && data instanceof Blob) {
        body = data;
    }
    else if (typeof ArrayBuffer !== 'undefined' && data instanceof ArrayBuffer) {
        body = data;
    }
    else if (typeof URLSearchParams !== 'undefined' && data instanceof URLSearchParams) {
        body = data;
    }
    else if (typeof ReadableStream !== 'undefined' && data instanceof ReadableStream) {
        body = data;
    }
    else { // json object or Array, or other
        // eslint-disable-next-line no-lonely-if
        if (args.processData && args.contentType && args.contentType.includes('json')) {
            body = JSON.stringify(data);
        }
        else if (args.processData) {
            body = QueryString.stringify(data);
        }
        else {
            body = data;
        }
    }
    return body;
}

/* eslint-disable @typescript-eslint/no-unsafe-return */
/**
 * fetch wrapper
 *
 * parameter init ignored during parameter input is typeof Request
 */
function _fetch(input, args, requestInit) {
    /* istanbul ignore else */
    if (!input) {
        throwError(() => new TypeError('value of input invalid'));
    }
    let req$ = createObbRequest(input, args, requestInit);
    req$ = parseRequestStream(req$, args);
    return req$.pipe(concatMap(res => handleRedirect(res, args, requestInit)));
}
/** Create Observable Request */
function createObbRequest(input, args, requestInit) {
    let inputNew = input;
    const fetchModule = selectFecthModule(args.fetchModule);
    if (typeof input === 'string') {
        if (['GET', 'DELETE'].includes(requestInit.method)) {
            inputNew = parseRequestGetLikeData(input, args);
        }
        else if (['POST', 'PUT', 'OPTIONS'].includes(requestInit.method)) {
            const body = parseRequestPostLikeData(args);
            if (typeof body !== 'undefined') {
                requestInit.body = body;
            }
        }
        else {
            // eslint-disable-next-line @typescript-eslint/restrict-template-expressions
            throw new TypeError(`Invalid method value: "${requestInit.method}"`);
        }
        return defer(() => fetchModule(inputNew, requestInit));
    }
    else {
        return defer(() => fetchModule(input));
    }
}
function parseRequestStream(request$, args) {
    const req$ = parseTimeout(request$, args.timeout, args.abortController);
    return req$;
}
function parseTimeout(request$, timeoutValue, abortController) {
    let ret$ = request$;
    /* istanbul ignore else */
    if (typeof timeoutValue === 'number' && timeoutValue >= 0) {
        ret$ = request$.pipe(timeout(timeoutValue), catchError((err) => {
            // test by test_browser/30_abort.test.ts
            // eslint-disable-next-line @typescript-eslint/unbound-method
            if (abortController && typeof abortController.abort === 'function' && !abortController.signal.aborted) {
                abortController.abort();
            }
            throw err;
        }));
    }
    return ret$;
}
/**
 * Handle redirect case to retrieve cookies before jumping under Node.js.
 * There's no effect under Browser
 *
 * docs: https://developer.mozilla.org/en-US/docs/Web/HTTP/Status
 */
function handleRedirect(resp, args, init) {
    // test by test/30_cookie.test.ts
    /* istanbul ignore next */
    if (args.keepRedirectCookies === true && resp.status >= 301 && resp.status <= 308) {
        const url = resp.headers.get('location');
        const cookie = resp.headers.get('Set-Cookie');
        /* istanbul ignore if */
        if (url) {
            const cookieObj = parseRespCookie(cookie);
            /* istanbul ignore else */
            if (cookieObj) {
                args.cookies = args.cookies
                    ? Object.assign(Object.assign({}, args.cookies), cookieObj) : Object.assign({}, cookieObj);
            }
            const options = parseInitOpts({ args, requestInit: init });
            if (resp.status === 303) {
                options.requestInit.method = 'GET';
                return _fetch(url, options.args, options.requestInit);
            }
            else {
                return _fetch(url, options.args, options.requestInit);
            }
        }
    }
    else {
        throwError(() => 'Redirect location is empty');
    }
    return of(resp);
}

/* eslint-disable @typescript-eslint/no-explicit-any */
/**
 * Observable fetch
 *
 * parameter init ignored during parameter input is typeof Request
 */
function rxfetch(input, init) {
    /* istanbul ignore else */
    if (!input) {
        throwError(() => new TypeError('value of input invalid'));
    }
    const initOpts = init ? Object.assign(Object.assign({}, initialRxRequestInit), init) : Object.assign({}, initialRxRequestInit);
    const options = splitInitArgs(initOpts);
    const { args, requestInit } = parseInitOpts(options);
    const dataType = args.dataType;
    const req$ = _fetch(input, args, requestInit);
    const ret$ = req$.pipe(concatMap(res => handleResponseError(res, dataType === 'bare')), concatMap(res => parseResponseType(res, dataType)));
    return ret$;
}
/** Observable GET method of fetch() */
function get(input, init) {
    const ps = init
        ? Object.assign(Object.assign({}, init), { method: 'GET' }) : { method: 'GET' };
    return rxfetch(input, ps);
}
/** Observable POST method of fetch() */
function post(input, init) {
    const ps = init
        ? Object.assign(Object.assign({}, init), { method: 'POST' }) : { method: 'POST' };
    return rxfetch(input, ps);
}
/** Observable PUT method of fetch() */
function put(input, init) {
    const ps = init
        ? Object.assign(Object.assign({}, init), { method: 'PUT' }) : { method: 'PUT' };
    return rxfetch(input, ps);
}
/** Observable DELETE method of fetch() */
function remove(input, init) {
    const ps = init
        ? Object.assign(Object.assign({}, init), { method: 'DELETE' }) : { method: 'DELETE' };
    return rxfetch(input, ps);
}
/** Update initialRxRequestInit */
function setGloalRequestInit(config) {
    for (const [key, value] of Object.entries(config)) {
        Object.defineProperty(initialRxRequestInit, key, {
            configurable: true,
            enumerable: true,
            writable: true,
            // eslint-disable-next-line @typescript-eslint/no-unsafe-assignment
            value,
        });
    }
}
/** Get copy of initialRxRequestInit */
function getGloalRequestInit() {
    return Object.assign({}, initialRxRequestInit);
}

export { ContentTypeList, buildQueryString, rxfetch as fetch, get, getGloalRequestInit, parseInitOpts, parseRequestGetLikeData, parseRequestPostLikeData, parseRespCookie, post, put, remove, rxfetch, selectFecthModule, setGloalRequestInit, splitInitArgs };
//# sourceMappingURL=rxxfetch.esm.js.map
